/*
 * Recording_pz2
 *
 * 2006 Jannis Achstetter (jannis),
 * improved and fixed 2007 Corbin Simpson
 *
 * Please don't blame me for lame, slow or ugly code,
 * I'm just starting to learn C. Send patches if you want
 * and I will learn and do better next time.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Further editing and formatting done by Corbin Simpson under the terms
 * of the GNU GPL, version 2 or later. Enjoy!
 */

/* Version: 0.4
 * Changes:
 *  * Clock becomes "Elapsed Time" when recording
 *  * Slight speedup in main recording loop
 *  * Recordings are now created with friendlier permissions (ext3/HFS+ iPods)
 *  * More code cleanup; compiles cleanly
 * Bugs:
 *  * Code still generates some warnings
 *  * Clipped beep and signal spike on start of recording (DSP driver issue?)
 *  * Negative offset on main recording body (DSP driver issue?)
 *  * No PB menu/PB
 * Requests & Features:
 *  * "Eye candy"
 *  * PB menu
 *
 * If you want, you can uncomment update_pb_menu() and have a laugh at my
 * ineptitude with C. Enjoy!
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include "pz.h"

//directory to read + write
#ifdef IPOD
#define RECORDINGS "/Recordings"
#else
#define RECORDINGS "Recordings"
#endif
//buffersizes
#define DSP_REC_SIZE	4*1024*2*2
#define DSP_PLAY_SIZE	4*1024*2*2
//modes
#define RECORD        0
#define PLAYBACK      1


/* --- VARS ---- */
static PzModule *module;	//opaque type, used only to fetch path info
static PzWindow *recwindow;	// The recording window
static PzWindow *pbwindow;
static PzWidget *recwid;	// The widget where stuff actually happens
static PzWidget *pbwid;
ttk_surface recsrf;
ttk_surface pbsrf;
static PzConfig *conf;


//sound
static pz_dsp_st dspz;
static int mode;
const char *srates[] =
  { "8000", "32000", "44100", "48000", "88200", "96000", 0 };
const char *rmodes[] = { "Mic", "Line In", 0 };
static int samplerate = 0, channels = 1, playing = 0, recording = 0, paused = 0, start_pos = 0, file_fd, filelength;	// Status descriptors
static char pcm_file[128];
static unsigned short pcm_buf[32 * 1024];
static long currenttime = 0;	// What's this used for?
static time_t start_rec_time;	// This is for elapsed time recording.
static DIR *pb_dir;


/* --- Prototypes --- */
void init_record ();
static int settings_button (TWidget * this, int key, int time);
PzWindow *new_rec_window ();
void draw_rec (PzWidget * wid, ttk_surface srf);
PzWindow *new_pb_window ();
void draw_pb (PzWidget * wid, ttk_surface srf);
int handle_event (PzEvent * ev);
void start_record ();
void end_record ();
unsigned int read32_le (int file_fd);
unsigned short read_16_le (int file_fd);
int write_32_le (int file_fd, unsigned int value);
int write_16_le (int file_fd, unsigned short value);
int write_wav_header (int file_fd, int samplerate, int channels);
void cleanup_record ();
void unload_record ();
//void update_pb_menu (); // Updates PB menu entries.

/* --- Implementation --- */
/********************************************************************/
/********************************************************************/

//INIT (is called at after pz2-startup)
void
init_record ()
{
  module = pz_register_module ("record", unload_record);
  conf = pz_load_config (pz_module_get_cfgpath (module, "record.conf"));
  pz_menu_add_action ("/Extras/Utilities/Recording/Record", new_rec_window);
  pz_menu_add_action ("/Extras/Utilities/Recording/Playback", new_pb_window);
  //update_pb_menu ();
  pz_menu_add_setting ("/Extras/Utilities/Recording/Settings/Rate", 1, conf,
		       srates);
  pz_menu_add_setting ("/Extras/Utilities/Recording/Settings/Mode", 2, conf,
		       rmodes);
  ((TWidget *) pz_get_menu_item ("/Extras/Utilities/Recording/Settings")->
   data)->button = settings_button;
}

/* void
update_pb_menu () {
  pb_dir = opendir(RECORDINGS);
  struct dirent *pb_file;
  char temp_string[100];
  while (pb_file = readdir(pb_dir)) {
    if (strcmp(pb_file->d_name[0], ".") == 0) {
      continue;
    }
    temp_string = "/Extras/Utilities/Recording/Playback/";
    strcat(temp_string, pb_file->d_name);
    pz_menu_add_action ();
  }
  closedir(pb_dir);
} */

static int
settings_button (TWidget * this, int key, int time)
{
  if (key == TTK_BUTTON_MENU)
    {
      pz_save_config (conf);
    }
  return ttk_menu_button (this, key, time);
}

PzWindow *
new_rec_window ()
{
  recwindow = pz_new_window ("Recording", PZ_WINDOW_NORMAL);
  recwid = pz_add_widget (recwindow, draw_rec, handle_event);
  pz_widget_set_timer (recwid, 200);
  mode = RECORD;
  recording = 0;
  return pz_finish_window (recwindow);
}

void
draw_rec (PzWidget * wid, ttk_surface srf)
{
  /* This draws the RECording window, not a RECtangle. */
  recsrf = srf;
  time_t t;
  char disptime[128];
  char text[128];
  char text2[128];
  struct tm *tm;

  t = time (NULL);
  /* If we're recording, let's update the display with time elapsed, not wall time. */
  if (recording)
    {
      t = (time_t) difftime (t, start_rec_time);
    }
  tm = localtime (&t);
  sprintf (disptime, "Time: %02d:%02d:%02d", tm->tm_hour, tm->tm_min,
	   tm->tm_sec);

  if (recording == 0)
    {
      sprintf (text, _("Press Play/Pause to start."));

      int y = wid->y + 10;
      ttk_text (srf, ttk_textfont,
		wid->x + (wid->w - ttk_text_width (ttk_textfont, text)) / 2,
		y, ttk_makecol (BLACK), text);
      y += 12;
      y += 20;
      ttk_text (srf, ttk_textfont,
		wid->x + (wid->w -
			  ttk_text_width (ttk_textfont, disptime)) / 2, y,
		ttk_makecol (BLACK), disptime);
      wid->dirty++;

    }
  else
    {
      //This was sprintf, but strcpy is faster here!
      strcpy (text, "Currently recording...");
      sprintf (text2, "Channels: %d Hz: %d", channels, samplerate);

      ssize_t n, rem;
      unsigned short *p = pcm_buf;
      rem = n = pz_dsp_read (&dspz, (void *) p, DSP_REC_SIZE *
			     (samplerate > 8000 ? 2 : 1));
      if (n > 0)
	{
	  while (rem)
	    {
	      n = write (file_fd, (void *) p, rem);
	      if (n > 0)
		{
		  rem -= n;
		  p += (n / 2);
		}
	    }
	}

      int y = wid->y + 10;
      ttk_text (srf, ttk_textfont,
		wid->x + (wid->w - ttk_text_width (ttk_textfont, text)) / 2,
		y, ttk_makecol (BLACK), text);
      y += 12;
      ttk_text (srf, ttk_textfont,
		wid->x + (wid->w - ttk_text_width (ttk_textfont, text2)) / 2,
		y, ttk_makecol (BLACK), text2);
      y += 20;
      ttk_text (srf, ttk_textfont,
		wid->x + (wid->w -
			  ttk_text_width (ttk_textfont, disptime)) / 2, y,
		ttk_makecol (BLACK), disptime);
      wid->dirty++;
    }

}

PzWindow *
new_pb_window ()
{
  pbwindow = pz_new_window ("Playback", PZ_WINDOW_NORMAL);
  pbwid = pz_add_widget (pbwindow, draw_pb, handle_event);
  pz_widget_set_timer (pbwid, 250);
  mode = PLAYBACK;
  return pz_finish_window (pbwindow);
}

void
draw_pb (PzWidget * wid, ttk_surface srf)
{
  pbsrf = srf;
  char *text = "Not implemented yet";
  int y = wid->y + 10;
  ttk_text (srf, ttk_textfont,
	    wid->x + (wid->w - ttk_text_width (ttk_textfont, text)) / 2, y,
	    ttk_makecol (BLACK), text);
  y += 12;
  text = "Please use 'play' so long";
  ttk_text (srf, ttk_textfont,
	    wid->x + (wid->w - ttk_text_width (ttk_textfont, text)) / 2, y,
	    ttk_makecol (BLACK), text);
  wid->dirty++;
}

int
handle_event (PzEvent * ev)
{
  switch (ev->type)
    {
    case PZ_EVENT_TIMER:
      if (ev->wid == recwid)
	draw_rec (ev->wid, recsrf);
      if (ev->wid == pbwid)
	draw_pb (ev->wid, pbsrf);
      break;
    case PZ_EVENT_BUTTON_UP:
      switch (ev->arg)
	{
	case PZ_BUTTON_MENU:
	  pz_close_window (ev->wid->win);
	  break;
	case PZ_BUTTON_PLAY:
	  if (ev->wid == recwid)
	    {
	      if (recording == 0)
		{
		  start_record ();
		}
	      else
		{
		  end_record ();
		};
	      ev->wid->dirty++;
	    }
	  break;
	}
    }
  return 0;
}

void
start_record ()
{
  time_t t;
  struct tm *tm;
  int ret;

  t = start_rec_time = time (NULL);
  tm = gmtime (&t);
  sprintf (pcm_file, "%s/%02d%02d%04d-%02d%02d%02d.wav", RECORDINGS,
	   tm->tm_mon + 1, tm->tm_mday, tm->tm_year + 1900, tm->tm_hour,
	   tm->tm_min, tm->tm_sec);

  if (pz_get_int_setting (conf, 2) == 1)
    {
      channels = 2;
    }				// else = default = 1

  switch (pz_get_int_setting (conf, 1))
    {
    case 0:
      samplerate = 8000;
      break;
    case 1:
      samplerate = 32000;
      break;
    case 2:
      samplerate = 44100;
      break;
    case 3:
      samplerate = 48000;
      break;
    case 4:
      samplerate = 88200;
      break;
    case 5:
      samplerate = 96000;
      break;
    }

  if (channels == 1)
    {
      ret = pz_dsp_open (&dspz, PZ_DSP_MIC);
    }
  else
    {
      ret = pz_dsp_open (&dspz, PZ_DSP_LINEIN);
    };
  if (ret < 0)
    {
      pz_perror ("/dev/dsp");
    }
  pz_dsp_setup (&dspz, channels, samplerate);
  /* This is 644 (instead of 400) in order to be read from the OS with less work
   * so just leave it like this, please! */
  file_fd =
    open ((char *) pcm_file, O_WRONLY | O_CREAT | O_TRUNC,
	  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (file_fd < 0)
    {
      pz_perror ((char *) pcm_file);
    }
  write_wav_header (file_fd, samplerate, channels);

  /* start of data chunk */
  start_pos = (int) lseek (file_fd, 0, SEEK_CUR);
  write (file_fd, "data", 4);
  write_32_le (file_fd, 0);	// dummy length value

  recording = 1;
}

void
end_record ()
{
  pz_dsp_close (&dspz);

  /* get file length by position which is 1 more than last */
  filelength = (int) lseek (file_fd, 1, SEEK_CUR) - 1;

  /* correct length value for the data chunk */
  lseek (file_fd, start_pos + 4, SEEK_SET);
  write_32_le (file_fd, filelength - (start_pos + 8));

  /* correct length value for the RIFF header */
  lseek (file_fd, 4, SEEK_SET);
  write_32_le (file_fd, filelength - 8);

  close (file_fd);

  recording = 0;
}

unsigned int
read_32_le (int file_fd)
{
  int value;
  read (file_fd, &value, 4);
  return value;
}

unsigned short
read_16_le (int file_fd)
{
  short value;
  read (file_fd, &value, 2);
  return value;
}

int
write_32_le (int file_fd, unsigned int value)
{
  return write (file_fd, &value, 4) == 4;
}

int
write_16_le (int file_fd, unsigned short value)
{
  return write (file_fd, &value, 2) == 2;
}

int
write_wav_header (int file_fd, int samplerate, int channels)
{
  write (file_fd, "RIFF", 4);
  write_32_le (file_fd, 0);	// remaining length after this header
  write (file_fd, "WAVE", 4);

  /* fmt chunck */
  write (file_fd, "fmt ", 4);
  write_32_le (file_fd, 16);	// length of fmt is 16

  write_16_le (file_fd, 1);	// WAVE_FORMAT_PCM
  write_16_le (file_fd, channels);	// no. channels
  write_32_le (file_fd, samplerate);	// sample rate
  write_32_le (file_fd, samplerate * channels * 2);	// avg bytes per second
  write_16_le (file_fd, channels * 2);	// block align
  write_16_le (file_fd, 16);	// bits per sample

  return 0;
}


//Module unloading (when exiting PZ)
void
unload_record ()
{
  pz_save_config (conf);
}

//INIT-function-caller
PZ_MOD_INIT (init_record)	//No semicolon at end of line
